﻿using Configuration;
using Microsoft.SPOT.Hardware;
using Oberon.HttpResources.Netmf;
using Oberon.HttpServers;
using System.Threading;

class ParallelBlinker
{
    static void Main()
    {
        Parameters.Setup();

        var buffer = new Buffer { };

        var blinker = new Blinker { SourceBuffer = buffer };

        var webServer = new HttpServer
        {
            StreamProvider = Parameters.StreamProvider,
            RequestRouting =
            {
                {
                    "PUT /blinkingPeriod/target",
                    new ManipulatedVariable
                    {
                        FromHttpRequest =
                            CSharpRepresentation.TryDeserializeInt,
                        ToActuator = buffer.HandlePut
                    }.HandleRequest
                },
                {
                    "GET /blinkingPeriod/target.html",
                    HandleBlinkTargetHtml
                }
            }
        };

        var blinkerThread = new Thread(blinker.Run);
        blinkerThread.Start();
        webServer.Run();
    }

    static void HandleBlinkTargetHtml(RequestHandlerContext context)
    {
        string requestUri =
            context.BuildRequestUri("/blinkingPeriod/target");
        var script =
            @"<html>
                <head>
                  <script type=""text/javascript"">
                    var r;
                    try {
                      r = new XMLHttpRequest();
                    } catch (e) {
                      r = new ActiveXObject('Microsoft.XMLHTTP');
                    }
                    function put (content) {
                      r.open('PUT', '" + requestUri + @"');
                      r.setRequestHeader(""Content-Type"", ""text/plain"");
                      r.send(document.getElementById(""period"").value);
                    }
                  </script>
                </head>
                <body>
                  <p>
                    <input type=""text"" value=""500"" id=""period"">
                    <input
                      type=""button"" value=""Set"" onclick=""put()""/>
                  </p>
                </body>
              </html>";
        context.SetResponse(script, "text/html");
    }

}

public class Blinker
{
    public Buffer SourceBuffer { get; set; }

    public void Run()
    {
        var ledPort = new OutputPort(Parameters.LedPin, false);
        var period = 500;
        var on = true;
        while (true)
        {
            Thread.Sleep(period / 2);

            object setpoint = SourceBuffer.HandleGet();
            if (setpoint != null)
            {
                period = (int)setpoint;
                period = period > 10000 ? 10000 : period;
                period = period < 20 ? 20 : period;
            }

            on = !on;
            ledPort.Write(on);
        }
    }
}

/// <summary>
/// The Buffer class provides a means for parallel actors to exchange
/// data in a thread-safe way. One actor puts objects into a buffer
/// instance, the other fetches them from the buffer.
/// This buffer class is able to buffer at most one object at a time:
/// if the producer puts a new object into the buffer, its old contents
/// is replaced by the new one. When the consumer gets the current
/// buffer contents, the buffer is not modified in any way.
/// This buffer class is intended for actors that are synchronized to
/// real-time ("time-triggered"); no mechanism for direct
/// synchronization between the actors is provided.
/// </summary>
public sealed class Buffer
{
    readonly static object Lock = new object();

    object buffer = null;

    /// <summary>
    /// Puts object o, which may be null, into the buffer.
    /// The current buffer contents is overwritten with the new one.
    /// It is safe to call this method from different threads.
    /// </summary>
    /// <param name="o">New value to be buffered.</param>
    public void HandlePut(object o)
    {
        lock (Lock)
        {
            buffer = o;
        }
    }

    /// <summary>
    /// Returns the object that is currently contained in the buffer,
    /// or null.
    /// It is safe to call this method from different threads.
    /// </summary>
    /// <returns>Currently buffered value. May be null.</returns>
    public object HandleGet()
    {
        object o;
        lock (Lock)
        {
            o = buffer;
        }
        return o;
    }
}
